// refcount.h

#ifndef _REFCOUNT_H_
#define _REFCOUNT_H_

//------------------------------------------------------------------------------
// RefCountBase
//------------------------------------------------------------------------------

class RefCountBase
{
public:
	RefCountBase()
		: ref_counter_(0)
	{}

	virtual ~RefCountBase()
	{}

public:
	int ref_counter_;
};

//------------------------------------------------------------------------------
// RefCountMix
//------------------------------------------------------------------------------

class RefCountMix
{
public:
  RefCountMix() : ref_count_ ( 0 ) {}
  virtual ~RefCountMix() {}

public:
  int ref_count_;

private:
	RefCountMix ( const RefCountMix & rhs );
	RefCountMix& operator= ( const RefCountMix & rhs );
};

//------------------------------------------------------------------------------
// SharedObject
//------------------------------------------------------------------------------

template < typename T >
class SharedObject : public T, public RefCountMix
{};

//------------------------------------------------------------------------------
// SharedObjectPtr
//------------------------------------------------------------------------------

template < typename T >
class SharedObjectPtr
{
public:
	SharedObjectPtr() : obj_(0)
	{}
	
	SharedObjectPtr ( T * obj ) : obj_(obj)
	{
		Lock();
	}
	
	SharedObjectPtr ( const SharedObjectPtr & rhs ) : obj_ ( rhs.obj_ )
	{
		Lock();
	}
	
	SharedObjectPtr & operator= ( const SharedObjectPtr & rhs )
	{
		if ( this == & rhs )
			return *this;

		Unlock();

		obj_ = rhs.obj_;

		Lock();

		return *this;
	}
	
	~SharedObjectPtr()
	{
		Unlock();
	}

	T * operator->()
	{
		return obj_;
	}

	const T * operator-> () const
	{
		return obj_;
	}

	operator T * ()
	{
		return *obj_;
	}

private:

	void inline Lock()
	{
		if ( obj_ )
			++((obj_)->ref_count_);
	}

	void inline Unlock()
	{
		if ( obj_ && ( ( --(obj_)->ref_count_) == 0 ) )
				delete obj_;
	}

private:
	T * obj_;
};

//------------------------------------------------------------------------------

#endif // _REFCOUNT_H_